#include <xc.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <flash.h>
#include <pic16f1459.h>

#define PRESCALER 5
#define REAL_DELAY PRESCALER - 2
#define ONE_SECOND (6000000 >> REAL_DELAY)

#define VAR_LEN 10
#define SIDEREAL_DAY 86164
#define PROG_START_ADDRESS  0x1F00
#define MOUNT_EQ6 0x00
#define MOUNT_HEQ5 0x01
#define MOUNT_EQ5 0x02
#define MOUNT_EQ3 0x03
#define MOUNT_GT 0x80
#define MOUNT_MF 0x81
#define MOUNT_114GT 0x82
#define MOUNT_DOB 0x90
#define MOUNT_GEEHALEL 0xF0
#define MOUNT_CUSTOM 0xFE
#define MOUNT_TYPE MOUNT_CUSTOM
#define MOUNT_VERSION 0x000001
#define VERSION (MOUNT_VERSION + (MOUNT_TYPE * 65536))
#define TOTAL_STEPS_RA 2949120 //(STEPS_RA * WORM_RATIO_RA)
#define STEPS_RA 20480 //(MOTOR_STEPS_RA * HISPEED_RATIO)
#define GUIDE_STEPS_RA 43825 //(SIDEREAL_DAY * ONE_SECOND / TOTAL_STEPS_RA)
#define TOTAL_STEPS_DEC 2949120 //(STEPS_DEC * WORM_RATIO_DEC)
#define STEPS_DEC 20480 //(MOTOR_STEPS_DEC * HISPEED_RATIO)
#define GUIDE_STEPS_DEC 43825 //(SIDEREAL_DAY * ONE_SECOND / TOTAL_STEPS_DEC)
#define HISPEED_RATIO 64
#define FORWARD 0
#define MODE_SLEW 1
#define SPEED_LOW 0
#define BACKWARD 2
#define MODE_GOTO 0
#define SPEED_HIGH 4
#define RA_DIR -1
#define DEC_DIR -1
#define TRMT_DELAY 0
#define CYCLE_INCREMENT 0x100

#pragma config FOSC = INTOSC
#pragma config WDTE = OFF
#pragma config PWRTE = OFF
#pragma config MCLRE = ON
#pragma config CP = OFF
#pragma config BOREN = OFF
#pragma config CLKOUTEN = OFF
#pragma config IESO = OFF
#pragma config FCMEN = OFF
#pragma config WRT = OFF
#pragma config CPUDIV = NOCLKDIV
#pragma config USBLSCLK = 48MHz
#pragma config PLLMULT = 3x
#pragma config PLLEN = ENABLED
#pragma config STVREN = OFF
#pragma config BORV = HI
#pragma config LPBOR = OFF
#pragma config LVP = ON

    enum SkywatcherCommand {
        Initialize                = 'F',
        InquireMotorBoardVersion  = 'e',
        InquireGridPerRevolution  = 'a',
        InquireTimerInterruptFreq = 'b',
        InquireHighSpeedRatio     = 'g',
        InquirePECPeriod          = 's',
        InstantAxisStop           = 'L',
        NotInstantAxisStop        = 'K',
        SetAxisPositionCmd        = 'E',
        GetAxisPosition           = 'j',
        GetAxisStatus             = 'f',
        SetSwitch                 = 'O',
        SetMotionMode             = 'G',
        SetGotoTargetIncrement    = 'H',
        SetBreakPointIncrement    = 'M',
        SetGotoTarget             = 'S',
        SetBreakStep              = 'U',
        SetStepPeriod             = 'I',
        StartMotion               = 'J',
        GetStepPeriod             = 'D', // See Merlin protocol http://www.papywizard.org/wiki/DevelopGuide
        ActivateMotor             = 'B', // See eq6direct implementation http://pierre.nerzic.free.fr/INDI/
        SetST4GuideRateCmd        = 'P',
        GetHomePosition           = 'd', // Get Home position encoder count (default at startup)
        SetFeatureCmd             = 'W', // EQ8/AZEQ6/AZEQ5 only
        GetFeatureCmd             = 'q', // EQ8/AZEQ6/AZEQ5 only
        InquireAuxEncoder         = 'd', // EQ8/AZEQ6/AZEQ5 only
        SetVars                   = '@',
      num_commands
    };
    enum axes {
        RA = 0,
        DEC = 1,
        num_axes
    };
    unsigned char sin_table[] = {
3,
6,
8,
11,
14,
17,
19,
21,
23,
25,
27,
28,
29,
30,
30,
31,
30,
30,
29,
28,
27,
25,
23,
21,
19,
17,
14,
11,
8,
6,
3,
0
    };
    signed short long trmt = 0;
    void loop();
    void send_string();
    void long2str(signed short long n, char *str);
    void SetMicrostep(unsigned char axis);
    void increment_decrement_target(unsigned char a, signed short long value);
    volatile unsigned char step;
    volatile unsigned char dir;
    int idx = 0;
    int xdi = 0;
    char str[3];
    char reply[7];
    char outstr[9];
    const unsigned short long customvars[VAR_LEN] @ (PROG_START_ADDRESS) = {TOTAL_STEPS_RA, TOTAL_STEPS_DEC, STEPS_RA, STEPS_DEC, VERSION, HISPEED_RATIO, RA_DIR, DEC_DIR, GUIDE_STEPS_RA, GUIDE_STEPS_DEC};
    signed long cycle[num_axes];
    unsigned long microstep[num_axes];
    unsigned long targetmicrostep[num_axes];
    unsigned short long pos[num_axes];
    unsigned short long target[num_axes];
    signed short long stepperiod[num_axes];
    signed short long breakpointincrement[num_axes];
    signed short long breaksteps[num_axes];
    signed short long breakstep[num_axes];
    signed short long usteps[num_axes];
    signed short long increment[num_axes];
    unsigned char stop[num_axes];
    unsigned char status[num_axes];
    unsigned char new_status[num_axes];
    unsigned char running[num_axes];
    unsigned char initialized[num_axes];
    unsigned char out_buf[11];
    signed long guidecycle[num_axes];
    signed short long guidestep[num_axes];
    
    unsigned char HEX2STR(unsigned char c)
    {
        return (((c)<0xA)?(c+'0'):(c-10)+'A');
    }
    unsigned char STR2HEX(char c)
    {
        return (((c)<'A')?(c-'0'):(c-'A')+10);
    }
int main(void)
{
	OSCCONbits.IRCF = 0b1111; /* 0b1111 = 16MHz HFINTOSC postscalar */
    OPTION_REGbits.PS = PRESCALER;
    OPTION_REGbits.T0CS = 0;
    OPTION_REGbits.PSA = 0;
    GIE = 1;
    PEIE = 1;
    TMR0 = 0x00;
    TMR0IE = 1;
    
    OPTION_REGbits.nWPUEN = 0;
    
    WPUA = 0x30;
    TRISA = 0x30;
    ANSELA = 0x00;
    PORTA = 0x00;
    
    WPUB = 0xF0;
    TRISB = 0xF0;
    ANSELB = 0x00;
    PORTB  = 0x00;
    
    TRISC = 0x00;
    ANSELC = 0x00;
    PORTC  = 0x00;

    TXSTAbits.BRGH = 0;
    SPBRG = 77;
    RCSTAbits.SPEN = 1;
    RCSTAbits.CREN = 1;
    TXSTAbits.TXEN = 1;
    PIE1bits.RCIE = 1;
    RCIE = 0;
    RC7 = 0;
    RC4 = 0;
    for(int axis=0; axis < num_axes; axis++)
    {
        pos[axis] = 0x800000;
        target[axis] = 0x800000;
        status[axis] = MODE_SLEW | FORWARD | SPEED_LOW;
        new_status[axis] = MODE_SLEW | FORWARD | SPEED_LOW;
        running[axis] = 0;
        initialized[axis] = 0;
        stepperiod[axis] = 0;
        guidestep[axis] = customvars[axis + 8];
    }
    loop();
}
                
void long2str(signed short long n, char *s) {
    s[1]=HEX2STR(n & 0xF); n >>= 4;
    s[0]=HEX2STR(n & 0xF); n >>= 4;
    s[3]=HEX2STR(n & 0xF); n >>= 4;
    s[2]=HEX2STR(n & 0xF); n >>= 4;
    s[5]=HEX2STR(n & 0xF); n >>= 4;
    s[4]=HEX2STR(n & 0xF);
    s[6]='\r';
    s[7] = '\0';
}

signed short long str2long(volatile char *s) {
    signed short long res = 0;
    res|=STR2HEX(s[4]); res <<= 4;
    res|=STR2HEX(s[5]); res <<= 4;
    res|=STR2HEX(s[2]); res <<= 4;
    res|=STR2HEX(s[3]); res <<= 4;
    res|=STR2HEX(s[0]); res <<= 4;
    res|=STR2HEX(s[1]);
    return res;
}

void send_string()
{
    if(trmt < TRMT_DELAY) {
        trmt ++;
    }
    else if(outstr[trmt - TRMT_DELAY] != 0 && TXSTAbits.TRMT) {
        TXREG = outstr[trmt - TRMT_DELAY];
        trmt++;
    }
}
void poll_chars()
{
    if(PIR1bits.RCIF)
    {
        PIR1bits.RCIF = 0;
        unsigned char c = RCREG;
        if(RCSTAbits.OERR)
        {
            RCSTAbits.CREN = 0;
            RCSTAbits.CREN = 1;
        }
        out_buf[idx++] = c;
        if(c == ':')
        {
            idx = 0;
        }
        if(c == '\r' || c == '\0')
        {
            char axis = out_buf[1] - '1';
            outstr[0] = '=';
            outstr[1] = '\r';
            outstr[2] = '\0';
            signed long arg;
            switch (out_buf[0])
            {
                case GetAxisPosition:
                    long2str(pos[axis], &outstr[1]);
                    break;
                case InquirePECPeriod:
                    long2str(0, &outstr[1]);
                    break;
                case SetST4GuideRateCmd:
                    guidestep[axis] = (out_buf[2] - '0') * customvars[axis + 8];
                    break;
                case SetBreakStep:
                case SetBreakPointIncrement:
                    break;
                case ActivateMotor:
                case Initialize:
                    initialized[axis] = 1;
                    break;
                case InquireMotorBoardVersion:
                    long2str(customvars[4], &outstr[1]);
                    break;
                case InquireGridPerRevolution:
                    long2str(customvars[axis], &outstr[1]);
                    break;
                case InquireTimerInterruptFreq:
                    long2str(customvars[axis + 2], &outstr[1]);
                    break;
                case InquireHighSpeedRatio:
                    long2str(customvars[5], &outstr[1]);
                    break;
                case SetAxisPositionCmd:
                    pos[axis] = str2long(&out_buf[2]);
                    target[axis] = pos[axis];
                    break;
                case SetMotionMode:
                    new_status[axis] = 0;
                    if(out_buf[2] == '1' || out_buf[2] == '3')
                        new_status[axis] |= MODE_SLEW;
                    if(out_buf[2] == '0' || out_buf[2] == '3')
                        new_status[axis] |= SPEED_HIGH;
                    if(out_buf[3] == '1')
                        new_status[axis] |= BACKWARD;
                    break;
                case GetAxisStatus:
                    sprintf(&outstr[1], "%c%c%c\r", status[axis] + '0', running[axis] + '0', initialized[axis] + '0');
                    break;
                case SetGotoTarget:
                    target[axis] = str2long(&out_buf[2]);
                    break;
                case SetGotoTargetIncrement:
                    increment[axis] = str2long(&out_buf[2]);
                    break;
                case SetStepPeriod:
                    stepperiod[axis] = str2long(&out_buf[2]);
                    break;
                case GetStepPeriod:
                    long2str(stepperiod[axis], &outstr[1]);
                    break;
                case StartMotion:
                    if(!(status[axis] & MODE_SLEW)) {
                        stepperiod[axis] = 0;
                        increment_decrement_target(axis, increment[axis]);
                    }
                    SetMicrostep(axis);
                    running[axis] = 1;
                    break;
                case SetVars:
                    arg = str2long(&out_buf[2]);
                    flash_write_code(PROG_START_ADDRESS + 3 * axis, &arg, 3);
                    break;
                case NotInstantAxisStop:
                case InstantAxisStop:
                    stop[axis] = 1;
                    break;
                default:
                    outstr[0] = '!';
                    break;
            }
            trmt = 0;
        }
    }
}

void increment_decrement_target(unsigned char a, signed short long value)
{
    target[a] += ((status[a] & BACKWARD) ? -value : value);
}
void guide(unsigned char a, signed short long value)
{
    pos[a] += customvars[6 + a] * value;
    target[a] += customvars[6 + a] * value;
}

void SetMicrostep(unsigned char a)
{
    stepperiod[a] = max(30, stepperiod[a]);
    breaksteps[a] = (SIDEREAL_DAY / (customvars[a] / customvars[a + 2])) / stepperiod[a];
    breakpointincrement[a] = customvars[a] / 10000;
    targetmicrostep[a] = stepperiod[a] * ONE_SECOND / (customvars[a + 2]) >> ((status[a] & SPEED_HIGH) ? 2 : 0);
    microstep[a] = ((status[a] & SPEED_HIGH) || !(status[a] & MODE_SLEW)) ? targetmicrostep[a] * (breaksteps[a] + 1) : targetmicrostep[a];
    breakstep[a] = breakpointincrement[a];
    stop[a] = 0;
}

void moveaxis(unsigned char chan, signed short long diff)
{
    unsigned char dir = (diff > 0);
    if(chan)
    {
        RC4 = dir;
    }
    else
    {
        RC7 = dir;
    }
}

void tmr_routine()
{
    if(INTCONbits.TMR0IF)
    {
        INTCONbits.TMR0IF = 0;
        if(guidecycle[0] >= guidestep[0] && !(status[0] & SPEED_HIGH))
        {
            guidecycle[0] = 0;
            if(!PORTAbits.RA4)
            {
                guide(0, 1);
            }
            else if(!PORTBbits.RB6)
            {
                guide(0, -1);
            }
        }
        if(guidecycle[1] >= guidestep[1] && !(status[1] & SPEED_HIGH))
        {
            guidecycle[1] = 0;
            if(!PORTAbits.RA5)
            {
                guide(1, 1);
            }
            else if(!PORTBbits.RB4)
            {
                guide(1, -1);
            }
        }
        for(int a = 0; a < num_axes; a++)
        {
            if(status[a] != new_status[a])
                status[a] = new_status[a];
            if(cycle[a] >= microstep[a])
            {
                cycle[a] = 0;
                int steps = 5;
                if(status[a]&SPEED_HIGH)
                    steps = 65;
                if(status[a] & MODE_SLEW) {
                    if(running[a])
                    {
                        increment_decrement_target(a, steps);
                    }
                }
                signed short long diff = (target[a] - pos[a]);
                if(!(status[a] & MODE_SLEW) && !stop[a]) {
                    signed short long d = (status[a] & BACKWARD) ? -diff : diff;
                    stop[a] = (d < min(increment[a] / 2, breakpointincrement[a] * breaksteps[a]));
                }
                if(breakstep[a] < 0 && ((status[a] & SPEED_HIGH) || !(status[a] & MODE_SLEW)))
                {
                    breakstep [a] = breakpointincrement[a];
                    if(microstep[a] > targetmicrostep[a] && !stop[a]) {
                        microstep[a] -= targetmicrostep[a];
                    }
                    else if(microstep[a] < (targetmicrostep[a] * breaksteps[a]) && stop[a]) {
                        microstep[a] += targetmicrostep[a] * 2;
                    }
                    else if(microstep[a] >= (targetmicrostep[a] * breaksteps[a]) && stop[a]) {
                        target[a] = pos[a];
                        running[a] = 0;
                        status[a] |= MODE_SLEW;
                        status[a] &= 0xff^SPEED_HIGH;
                    }
                }
                if(!(status[a] & SPEED_HIGH) && (status[a] & MODE_SLEW) && stop[a]) {
                    target[a] = pos[a];
                    running[a] = 0;
                }
                if(running[a])
                {
                    steps = (diff != 0) ? steps : 0;
                    pos[a] += (diff < 0) ? -steps : steps;
                    breakstep [a] -= steps;
                }
            }
            cycle[a] += CYCLE_INCREMENT;
        }
        guidecycle[0] += CYCLE_INCREMENT;
        guidecycle[1] += CYCLE_INCREMENT;
    }
}

void loop()
{
    while(1)
    {
        for(int a = 0; a < num_axes; a++)
        {
            send_string();
            if(initialized[a])
            {
                signed short long p = (customvars[a + 6] > 0 ? pos[a] : (customvars[a + 6] < 0 ? (0x7fffff - pos[a]) : 0)) / ((status[a] & SPEED_HIGH) ? 2 : 8);
                unsigned char port = PORTC;
                if(!(status[a] & SPEED_HIGH))
                {
                    usteps[a] ++;
                    if(usteps[a] >= sin_table[p % 32])
                    {
                        port &= 0x5f5 >> (a * 4);
                    }
                    else
                    {
                        if((p % 64) & 32)
                        {
                            port |= 0x02 << (a * 4);
                        }
                        else
                        {
                            port |= 0x08 << (a * 4);
                        }
                    }
                    if(usteps[a] >= sin_table[(p + 16) % 32])
                    {
                        port &= 0xafa >> (a * 4);
                    }
                    else
                    {
                        if(((p + 16) % 64) & 32)
                        {
                            port |= 0x01 << (a * 4);
                        }
                        else
                        {
                            port |= 0x04 << (a * 4);
                        }
                    }
                    if(usteps[a] >= 32)
                        usteps[a] = 0;
                }
                else
                {
                    port &= (0xf0 >> (a * 4));
                    port |= (((3 << (p % 4))&0xf)|(((p % 4) == 3) ? 1 : 0)) << (a * 4);
                }
                PORTC = port;
            }
        }
    }
}

void interrupt isr()
{
    poll_chars();
    tmr_routine();
}
